/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.

---
Copyright (C) 2011, Simon Hampe <hampe@mathematik.uni-kl.de>

This file provides c++ realizations of basic operations on polyhedral complexes
*/

#include "polymake/client.h"
#include "polymake/Rational.h"
#include "polymake/Matrix.h"
#include "polymake/Vector.h"
#include "polymake/linalg.h"
#include "polymake/IncidenceMatrix.h"
#include "polymake/atint/converters.h"
#include "polymake/atint/LoggingPrinter.h"
#include "polymake/atint/specialvarieties.h"
#include "polymake/atint/divisor.h"
#include "polymake/atint/WeightedComplexRules.h"

namespace polymake { namespace atint{ 
  
  using namespace atintlog::donotlog;
  //using namespace atintlog::dolog;
//   using namespace atintlog::dotrace;
  
  ///////////////////////////////////////////////////////////////////////////////////////
  
  //Documentation see header
  inline std::pair<Set<int>, Set<int> > separateRays(Matrix<Rational> m, Set<int> &affine, Set<int> &directional, bool uses_homog) {
    affine.clear();
    directional.clear();
    if(!uses_homog) {
      if(m.rows() > 0) directional = sequence(0,m.rows());
      //We use the dummy index -1 to indicate that a homogenized fan has just the origin as affine ray
      affine += -1;
      return std::pair<Set<int>,Set<int> >(affine,directional);
    }
    
    for(int row = 0; row < m.rows(); row++) {
      if(m(row,0) == 0) {
	directional += row;
      }
      else {
	affine += row;
      }
    }
    return std::pair<Set<int>, Set<int> >(affine,directional);
  }
  
  ///////////////////////////////////////////////////////////////////////////////////////
  
  //Documentation see header
  perl::Object compute_product_complex(const Array<perl::Object> &complexes) {
      //dbgtrace << "Generating container variables for result" << endl;
    
      //** EXTRACT FIRST COMPLEX ********************************************
      
      perl::Object firstComplex = complexes[0];
      //This will contain the sets describing the maximal cones
      IncidenceMatrix<> maximalConesInc = firstComplex.give("MAXIMAL_CONES");
      Vector<Set<int> > maximalCones = incmatrixToVector(maximalConesInc);
      //Will contain the rays
      Matrix<Rational> rayMatrix = firstComplex.give("RAYS");
      //Will contain the lineality space
      Matrix<Rational> linMatrix = firstComplex.give("LINEALITY_SPACE");
      //Will contain the weights (if any)
      Vector<Integer> weights;
      bool product_has_weights = false;
      if(firstComplex.lookup("TROPICAL_WEIGHTS") >> weights) {
	product_has_weights = true;
      }
      bool product_uses_homog = firstComplex.give("USES_HOMOGENEOUS_C");
      IncidenceMatrix<> local_restrictionInc = firstComplex.give("LOCAL_RESTRICTION");
      Vector<Set<int> > local_restriction = incmatrixToVector(local_restrictionInc);
      bool product_unimodular = firstComplex.give("IS_UNIMODULAR");
      bool product_has_lattice = firstComplex.exists("LATTICE_BASES");
      Matrix<Integer> product_l_generators;
      Vector<Set<int> > product_l_bases;
      if(product_has_lattice) {
	//dbgtrace << "Extracting first lattice " << endl;
	Matrix<Integer> lg = firstComplex.give("LATTICE_GENERATORS");
	  product_l_generators = lg;
	IncidenceMatrix<> lb = firstComplex.give("LATTICE_BASES");
	  product_l_bases = incmatrixToVector( lb);
      }
      
      //int product_dim = rayMatrix.cols() > linMatrix.cols() ? rayMatrix.cols() : linMatrix.cols();
      int product_dim = rayMatrix.rows() > 0? rayMatrix.cols() : linMatrix.cols();
      //Sort rays by affine and directional
      Set<int> product_affine;
      Set<int> product_directional;
      separateRays(rayMatrix, product_affine, product_directional, product_uses_homog);
	//dbgtrace << "Product affine " << product_affine << endl;
	//dbgtrace << "Product directional " << product_directional << endl;
      
      
      //dbgtrace << "Iterating over all " << complexes.size() -1 << " complexes: " << endl;
      
      //** ITERATE OTHER COMPLEXES ********************************************
      
      for(unsigned int i = 1; i < complexes.size(); i++) {
	//dbgtrace << "Considering complex nr. " << i+1 << endl;
	//Extract properties
	
	bool uses_homog = complexes[i].give("USES_HOMOGENEOUS_C");
	bool uses_weights = false;
	Matrix<Rational> prerays = complexes[i].give("RAYS");
	Matrix<Rational> prelin = complexes[i].give("LINEALITY_SPACE");
	IncidenceMatrix<> premax = complexes[i].give("MAXIMAL_CONES");
	IncidenceMatrix<> pre_local_restrictionInc = complexes[i].give("LOCAL_RESTRICTION");
	Vector<Set<int> > pre_local_restriction = incmatrixToVector(pre_local_restrictionInc); 
	
	Array<Integer> preweights;
	if(complexes[i].exists("TROPICAL_WEIGHTS")) {
	  uses_weights = true;
	  preweights = complexes[i].give("TROPICAL_WEIGHTS");
	}
	
	// ** RECOMPUTE RAY DATA ***********************************************
	
	//Sort rays
	Set<int> complex_affine;
	Set<int> complex_directional;
	separateRays(prerays,complex_affine, complex_directional,uses_homog);
	  //dbgtrace << "Affine: " << complex_affine << endl;
	  //dbgtrace << "Directional: " << complex_directional << endl;
	//If this fan uses homog. coordinates, strip away the first column of rays and linear space
	if(uses_homog) {
	  if(prerays.rows() > 0) prerays = prerays.minor(All,~scalar2set(0));
	  if(prelin.rows() > 0) prelin = prelin.minor(All,~scalar2set(0));
	  //dbgtrace << "prerays: " << prerays << endl;
	  //dbgtrace << "rcols: " << prerays.cols() << endl;
	  //dbgtrace << "lcols: " << prelin.cols() << endl;
	}
	//int dim = prerays.cols() > prelin.cols() ? prerays.cols() : prelin.cols();
	int dim = prerays.rows() > 0? prerays.cols() : prelin.cols();
	//dbgtrace << "dim: " << dim << endl;
		
	//dbgtrace << "Creating ray matrix" << endl;
// 	dbgtrace << "Affine rays of product are " << rayMatrix.minor(product_affine,All) << endl;
// 	dbgtrace << "Affine rays of complex are " << prerays.minor(complex_affine,All) << endl;
	
	//Create new ray matrix
	Matrix<Rational> newRays(0,product_dim + dim);
	//First create affine rays
// 	Map<std::pair<int,int>,int> affineIndices; //For index conversion
	Map<int, Map<int,int> > affineIndices;
	if(product_uses_homog || uses_homog) {
	  for(Entire<Set<int> >::iterator prays = entire(product_affine); !prays.at_end(); prays++)  {
	      affineIndices[*prays] = Map<int,int>();
	      Vector<Rational> pRay;
	      if(*prays >= 0) pRay = rayMatrix.row(*prays);
	      else pRay = zero_vector<Rational>(product_dim);
	      for(Entire<Set<int> >::iterator crays = entire(complex_affine); !crays.at_end(); crays++) {
		Vector<Rational> cRay;
		if(*crays >= 0) cRay = prerays.row(*crays);
		else cRay = zero_vector<Rational>(dim);
		newRays = newRays / (pRay | cRay);
		affineIndices[*prays][*crays] = newRays.rows()-1;
// 		affineIndices(*prays,*crays) = newRays.rows()-1;
	      }
	  }
	}
	Set<int> newAffine = sequence(0, newRays.rows());
	
	//dbgtrace << "New affine rays read " << newRays << endl;
	
	//If the product is non-homog. and the complex is homog., we have to add a column of ones in front
	//We then also add a zero column in front of the old product rays for the directional rays
	if(!product_uses_homog && uses_homog && newRays.rows() > 0) {
	  newRays = ones_vector<Rational>(newRays.rows()) | newRays;
	  product_dim++;
	  if(rayMatrix.rows() > 0) rayMatrix = zero_vector<Rational>(rayMatrix.rows()) | rayMatrix;
	  if(linMatrix.rows() > 0) linMatrix = zero_vector<Rational>(linMatrix.rows()) | linMatrix;
	}
	
	//dbgtrace << "Adding directional rays" << endl;
	//Now add the directional rays of both cones
	Map<int,int> pdirIndices;
	Map<int,int> cdirIndices; //For index conversion
	Vector<Rational> product_zero = zero_vector<Rational>(product_dim);
	Vector<Rational> complex_zero = zero_vector<Rational>(dim);
	for(Entire<Set<int> >::iterator prays = entire(product_directional); !prays.at_end(); prays++)  {
	  newRays = newRays / (rayMatrix.row(*prays) | complex_zero);
	  pdirIndices[*prays] = newRays.rows()-1;
	}
	for(Entire<Set<int> >::iterator crays = entire(complex_directional); !crays.at_end(); crays++) {
	  newRays = newRays / (product_zero | prerays.row(*crays));
	  cdirIndices[*crays] = newRays.rows()-1;
	}
	Set<int> newDirectional = sequence(newAffine.size(),product_directional.size() + complex_directional.size());
	
	//dbgtrace << "Creating lineality matrix" << endl;
	
	//Create new lineality matrix
	if(prelin.rows() > 0) {
	  prelin = Matrix<Rational>(prelin.rows(),product_dim) | prelin;
	}
	if(linMatrix.rows() > 0) {
	  linMatrix = linMatrix | Matrix<Rational>(linMatrix.rows(), dim);
	}
	
	//dbgtrace << "Prelin = " << prelin << "\nlinMatrix = " << linMatrix << endl;
	
	// ** RECOMPUTE LATTICE DATA ***************************************
	
	//Compute lattice data
	bool complex_unimodular = complexes[i].give("IS_UNIMODULAR");
	product_unimodular = product_unimodular && complex_unimodular;
	bool complex_has_lattice = complexes[i].exists("LATTICE_BASES");
	product_has_lattice = product_has_lattice && complex_has_lattice;
	int lattice_index_translation = 0; //Number of row where new lattice gens. begin
	
	Vector<Set<int> > new_lattice_bases;
	
	Matrix<Integer> complex_lg;
	IncidenceMatrix<> complex_lb;
	if(product_has_lattice) {
	  Matrix<Integer> clg = complexes[i].give("LATTICE_GENERATORS");
	    complex_lg = clg;
	  IncidenceMatrix<> clb = complexes[i].give("LATTICE_BASES");
	    complex_lb = clb;
	  //Compute cartesian product of lattice matrices:
	  //Adjust dimension, then concatenate
	  if(uses_homog) {
	      if(complex_lg.rows() > 0) complex_lg = complex_lg.minor(All,~scalar2set(0));
	      if(!product_uses_homog) {
		product_l_generators = zero_vector<Integer>() | product_l_generators;
	      }
	  }
	  product_l_generators = 
	    product_l_generators | Matrix<Integer>(product_l_generators.rows(), dim);
	  complex_lg = 
	    Matrix<Integer>(complex_lg.rows(), product_dim) | complex_lg;	    
	  lattice_index_translation = product_l_generators.rows();
	  product_l_generators /= complex_lg;
	  
	}
	
	// ** RECOMPUTE CONES *******************************************
	
	
		
	//dbgtrace << "Creating cones" << endl;
	
 	//dbgtrace << "Ray matrix is " << newRays << endl;
 	//dbgtrace << "Affine indices " << affineIndices << endl;
 	//dbgtrace << "Directional indices product" << pdirIndices << endl;
 	//dbgtrace << "Directional indices complex" << cdirIndices << endl;
	
	//Now create the new cones and weights:
	Vector<Set<int> > newMaxCones;
	Vector<Integer> newWeights;
	//Make sure, we have at least one "cone" in each fan, even if it is empty
	if(maximalCones.dim() == 0) { maximalCones |= Set<int>();}
	if(premax.rows() == 0) { premax = premax / Set<int>();}
	for(int pmax = 0; pmax < maximalCones.dim(); pmax++) {
	    Set<int> product_cone = maximalCones[pmax];
	    if(!product_uses_homog && uses_homog) {
	      product_cone = product_cone + (-1);
	    }
 	    //dbgtrace << "Product cone: " << product_cone << endl;
	    for(int cmax = 0; cmax < premax.rows(); cmax++) {
	      Set<int> complex_cone = premax.row(cmax);
	      if(!uses_homog && product_uses_homog) {
		complex_cone = complex_cone + (-1);
	      }
 	      //dbgtrace << "Complex cone: " << complex_cone << endl;
	      Set<int> newcone;
	      Set<int> pAffine = product_cone * product_affine;
	      Set<int> pDirectional = product_cone * product_directional;
	      Set<int> cAffine = complex_cone * complex_affine;
	      Set<int> cDirectional = complex_cone * complex_directional;
	      
	      //First add the affine rays: For each pair of affine rays add the corresponding index from
	      //affineIndices
	      for(Entire<Set<int> >::iterator pa = entire(pAffine); !pa.at_end(); pa++) {
		for(Entire<Set<int> >::iterator ca = entire(cAffine); !ca.at_end(); ca++) {
// 		    if(cmax + pmax == 0) dbgtrace << *pa << "," << *ca << ": " << affineIndices[std::make_pair(*pa,*ca)] << endl;
		    newcone = newcone + affineIndices[*pa][*ca];
		}		
	      }
	      //Now add the directional indices
	      for(Entire<Set<int> >::iterator pd = entire(pDirectional); !pd.at_end(); pd++) {
		newcone = newcone + pdirIndices[*pd];
	      }
	      for(Entire<Set<int> >::iterator cd = entire(cDirectional); !cd.at_end(); cd++) {
		newcone = newcone + cdirIndices[*cd];
	      }
	      //dbgtrace << "Result: " << newcone << endl;
	      newMaxCones |= newcone;
	      //Compute weight
	      if(product_has_weights || uses_weights) {
		newWeights = newWeights | (product_has_weights? weights[pmax] : Integer(1)) * (uses_weights? preweights[cmax] : Integer(1));
	      }
	      //Compute lattice data
	      if(product_has_lattice) {
		Set<int> cone_l_basis = product_l_bases[pmax] + Set<int>( translate(complex_lb.row(cmax),lattice_index_translation));
		new_lattice_bases |= cone_l_basis;
	      }
	    }
	}
	
 	//dbgtrace << "Maximal cones now: " << newMaxCones << endl;
	
	//Compute the cross product of the local_restrictions
	Vector<Set<int> > new_local_restriction;
	if(local_restriction.dim() > 0 || pre_local_restriction.dim() > 0) {
	  //If one variety is not local, we take all its cones for the product
	  Vector<Set<int> > product_locality(local_restriction);
	  if(product_locality.dim() == 0) {
	    IncidenceMatrix<> prolocInc = computeAllCones(rayMatrix , maximalCones, product_uses_homog);
	    for(int plr = 0; plr < prolocInc.rows(); plr++) { product_locality |= prolocInc.row(plr); }
// 	    for(Entire<Set<int> >::iterator aRay = entire(product_affine); !aRay.at_end(); aRay++) {
// 	      Set<int> single; single += *aRay;
// 	      product_locality |= single;
// 	    }
	  }
	  Vector<Set<int> > pre_locality(pre_local_restriction);
	  if(pre_locality.dim() == 0) {
	    IncidenceMatrix<> prelocInc = computeAllCones(prerays, premax, uses_homog);
	    for(int per = 0; per < prelocInc.rows(); per++) { pre_locality |= prelocInc.row(per); }
// 	    for(Entire<Set<int> >::iterator aRay = entire(complex_affine); !aRay.at_end(); aRay++) {
// 	      Set<int> single; single += *aRay;
// 	      pre_locality |= single;
// 	    }
	  }
	  
	  //dbgtrace << "pro_locality " << product_locality << endl;
	  //dbgtrace << "pre_locality " << pre_locality << endl;
	  
	  for(int i = 0; i < product_locality.dim(); i++) {
	    Set<int> pAffine = product_locality[i] * product_affine;
	    if(!product_uses_homog && uses_homog) pAffine += (-1);
	    Set<int> pDirectional = product_locality[i] * product_directional;
	    for(int j = 0; j < pre_locality.dim(); j++) {
	      Set<int> local_cone;
	      Set<int> cAffine = pre_locality[j] * complex_affine;
	      if(product_uses_homog && !uses_homog) cAffine += (-1);
	      Set<int> cDirectional = pre_locality[j] * complex_directional;
	      //First add the affine rays: For each pair of affine rays add the corresponding index from
	      //affineIndices
	      for(Entire<Set<int> >::iterator pa = entire(pAffine); !pa.at_end(); pa++) {
		for(Entire<Set<int> >::iterator ca = entire(cAffine); !ca.at_end(); ca++) {
		    local_cone = local_cone + affineIndices[*pa][*ca];
		}		
	      }
	      //Now add the directional indices
	      for(Entire<Set<int> >::iterator pd = entire(pDirectional); !pd.at_end(); pd++) {
		local_cone = local_cone + pdirIndices[*pd];
	      }
	      for(Entire<Set<int> >::iterator cd = entire(cDirectional); !cd.at_end(); cd++) {
		local_cone = local_cone + cdirIndices[*cd];
	      }
	      new_local_restriction |=  local_cone;
	    }
	  }
	}
	
	// ** COPY VALUES ONTO NEW PRODUCT ***************************************
	
	//Copy values
	rayMatrix = newRays;
	linMatrix = linMatrix.rows() == 0? prelin : (prelin.rows() == 0? linMatrix : linMatrix / prelin);
	product_dim = rayMatrix.cols() > linMatrix.cols() ? rayMatrix.cols() : linMatrix.cols();
	product_affine = newAffine;
	product_directional = newDirectional;
	maximalCones = newMaxCones;
	weights = newWeights;
	product_uses_homog = product_uses_homog || uses_homog;
	product_has_weights = product_has_weights ||  uses_weights;
	local_restriction = Vector<Set<int> > (new_local_restriction);
	product_l_bases = new_lattice_bases;
      }
    
      //Fill fan with result
      perl::Object result("WeightedComplex");
	result.take("RAYS") << rayMatrix;
	result.take("LINEALITY_SPACE") << linMatrix;
	result.take("MAXIMAL_CONES") << maximalCones;
	if(product_has_weights) result.take("TROPICAL_WEIGHTS") << weights;
	result.take("USES_HOMOGENEOUS_C") << product_uses_homog;
	result.take("LOCAL_RESTRICTION") << local_restriction;
	if(product_has_lattice) {
	    result.take("LATTICE_BASES") << product_l_bases;
	    result.take("LATTICE_GENERATORS") << product_l_generators;
	}
	result.take("IS_UNIMODULAR") << product_unimodular;
	
      return result;
         
  }
  
  ///////////////////////////////////////////////////////////////////////////////////////
  
  //Documentation see header
  perl::Object compute_product_complex_lattice(const Array<perl::Object> &complexes) {
    for(unsigned int i = 0; i < complexes.size(); i++) {
      complexes[i].give("LATTICE_BASES");
    }
    return compute_product_complex(complexes);
  }
  
  ///////////////////////////////////////////////////////////////////////////////////////
  
  /**
   @brief Takes the ray matrix of a complex and returns two sets, describing which rows are affine and which are directional rays
   @param Matrix<Rational> m The ray matrix of the complex
   @param bool uses_homog Whether the complex uses homogeneous coordinates (if not, the result is trivial)
   @return perl::ListReturn A list containing two sets. In this order: Affine rays (vertices), directional rays   
   */
  perl::ListReturn separateRayMatrix(Matrix<Rational> m, bool uses_homog) {
    Set<int> affine, directional;
    separateRays(m,affine, directional,uses_homog);
    if(!uses_homog) {
      affine = Set<int>();
    }
    perl::ListReturn result;
      result << affine;
      result << directional;
    return result;
  }
  
  ///////////////////////////////////////////////////////////////////////////////////////
  
  //Documentation see header
  perl::Object facetRefinement(perl::Object fan, Matrix<Rational> facets) {
    for(int r = 0; r < facets.rows(); r++) {
      fan = intersect_container(fan,halfspace_complex(0,facets.row(r)));
    }
    return fan;
  }
  
  ///////////////////////////////////////////////////////////////////////////////////////
  
  //Documentation see header
  perl::ListReturn fan_decomposition(perl::Object complex) {
    //Extract values
    bool uses_homog = complex.give("USES_HOMOGENEOUS_C");
    if(!uses_homog) {
      perl::ListReturn singlefan;
	singlefan << complex;
      return singlefan;
    }
    Matrix<Rational> rays = complex.give("RAYS");
    IncidenceMatrix<> cones = complex.give("MAXIMAL_CONES");
      IncidenceMatrix<> verticesInCones = T(cones);
    Vector<Integer> weights = complex.give("TROPICAL_WEIGHTS");
    Set<int> affine = complex.give("VERTICES");
    Set<int> directional = complex.give("DIRECTIONAL_RAYS");
    int ambient_dim = complex.give("CMPLX_AMBIENT_DIM");
    IncidenceMatrix<> local_restrictionInc = complex.give("LOCAL_RESTRICTION");
    Vector<Set<int> > local_restriction = incmatrixToVector(local_restrictionInc);
    
    //Take only compatible vertices
    if(local_restriction.dim() > 0) {
//       affine *= accumulate(local_restriction,operations::add());
      Set<int> compatible;
      for(Entire<Set<int> >::iterator af = entire(affine); !af.at_end(); af++) {
	Set<int> single; single += *af;
	if(is_coneset_compatible(single, local_restriction)) {
	    compatible += *af;
	}
      }
      affine = compatible;
    }
    
    
    //Now go through all affine rays
    perl::ListReturn result;
    for(Entire<Set<int> >::iterator v = entire(affine); !v.at_end(); v++) {
      //For each cone containing this vertex, we get a cone of the Star at v
      Set<int> conesatv = verticesInCones.row(*v);
      //Whether a ray has been added to the Star (for an affine ray, this actually means the distance
      Vector<bool> hasBeenAdded(rays.rows()); 
      //Contains the new row indices of the rays
      Map<int,int> newIndex;
      Matrix<Rational> starRays(0,ambient_dim+1);
      Vector<Set<int> > starCones;
      Vector<Integer> starWeights;
      //Now transform all the cones containing v
      for(Entire<Set<int> >::iterator catv = entire(conesatv); !catv.at_end(); catv++) {
	starWeights |= weights[*catv];
	Set<int> otherrays = cones.row(*catv) - *v;
	Set<int> newcone;
	for(Entire<Set<int> >::iterator r = entire(otherrays); !r.at_end(); r++) {
	    //If the ray is already in the matrix, just add its new index
	    if(hasBeenAdded[*r]) {
	      newcone += newIndex[*r];
	    }
	    else {
	      hasBeenAdded[*r] = true;
	      newIndex[*r] = starRays.rows();
	      newcone += starRays.rows();
	      //A directional ray is just copied, for an affine one we take the difference to v
	      if(directional.contains(*r)) {
		starRays /= rays.row(*r);
	      }
	      else {
		starRays /= (rays.row(*r) - rays.row(*v));
	      }
	    }
	}
	starCones |= newcone;
      } //End iterate all cones at v
      starRays = starRays.minor(All,~scalar2set(0));
      perl::Object fan("WeightedComplex");
	fan.take("INPUT_RAYS") << starRays;
	fan.take("INPUT_CONES") << starCones;
	fan.take("TROPICAL_WEIGHTS") << starWeights;
	fan.take("USES_HOMOGENEOUS_C") << false;
      result << fan;
    }//End iterate all vertices
    
    return result;
  }
  
  ///////////////////////////////////////////////////////////////////////////////////////
  
  //Documentation see header
  perl::Object affineTransformation(perl::Object complex, Vector<Rational> translate, Matrix<Integer> matrix) {
    //Homogenize first
    bool uses_homog = complex.give("USES_HOMOGENEOUS_C");
    if(!uses_homog) complex = complex.CallPolymakeMethod("homogenize");
    matrix = zero_vector<Integer>() / matrix;
    matrix = unit_vector<Integer>(matrix.rows(),0) | matrix;
    translate = Rational(0) | translate;
    
    //Extract values
    Matrix<Rational> rays = complex.give("RAYS");
    Set<int> affine = complex.give("VERTICES");
    Matrix<Rational> linspace = complex.give("LINEALITY_SPACE");
    IncidenceMatrix<> cones = complex.give("MAXIMAL_CONES");
    Vector<Integer> weights = complex.give("TROPICAL_WEIGHTS");
    IncidenceMatrix<> local_restrictionInc = complex.give("LOCAL_RESTRICTION");
    Vector<Set<int> > local_restriction = incmatrixToVector(local_restrictionInc);
    
    //Transform rays and lin space
    linspace = linspace * matrix;
    rays = rays * matrix;
    for(Entire<Set<int> >::iterator r = entire(affine); !r.at_end(); r++) {
      rays.row(*r) = rays.row(*r) + translate;
    }
    
    perl::Object result("WeightedComplex");
      result.take("RAYS") << rays;
      result.take("MAXIMAL_CONES") << cones;
      result.take("TROPICAL_WEIGHTS") << weights;
      result.take("LINEALITY_SPACE") << linspace;
      result.take("USES_HOMOGENEOUS_C") << true;
      result.take("LOCAL_RESTRICTION") << local_restriction;
    return result;
    
  }

  ///////////////////////////////////////////////////////////////////////////////////////
  
  //Documentation see header
  perl::Object skeleton_complex(perl::Object complex, int k, bool preserve = false) {
    //Extract properties
    int cmplx_dim = complex.give("CMPLX_DIM");
    Matrix<Rational> rays = complex.give("RAYS");
    IncidenceMatrix<> maximalCones = complex.give("MAXIMAL_CONES");
    bool uses_homog = complex.give("USES_HOMOGENEOUS_C");
    Matrix<Rational> lineality = complex.give("LINEALITY_SPACE");
    int lineality_dim = complex.give("LINEALITY_DIM");
    IncidenceMatrix<> local_restrictionInc = complex.give("LOCAL_RESTRICTION");
    Vector<Set<int> > local_restriction = incmatrixToVector(local_restrictionInc);
    
    //If the skeleton dimension is too small, return the 0-cycle
    if(k < (uses_homog? 0 : 1) || k < lineality_dim) {
      return CallPolymakeFunction("zero_cycle");
    }
    
    //If the skeleton dimension is the fans dimension, return the fan
    if(k == cmplx_dim) {
      return complex;
    }
    
    Vector<Set<int> > new_local_restriction;
    
    //Now we compute the codimension one skeleton of the fan (cmplx_dim - k) times 
    IncidenceMatrix<> newMaximalCones = maximalCones;
    for(int i = 1; i <= (cmplx_dim - k); i++) {
      newMaximalCones = calculateCodimOneData(rays,newMaximalCones, uses_homog, lineality, local_restriction).codimOneCones;
    }
    
    //Now return the result - made irredundant, if preserve is false
    Matrix<Rational> newrays = rays;
    if(!preserve) {
      //Take the union of all cones to see what rays are used
      Set<int> usedRays;
      for(int c = 0; c < newMaximalCones.rows(); c++) {
	usedRays += newMaximalCones.row(c);
      }
      
      if(local_restriction.dim() > 0) {
	  Map<int,int> index_map; //Maps indices of old rays to indices of new rays
	  int newIndex = 0;
	  for(Entire<Set<int> >::iterator uR = entire(usedRays); !uR.at_end(); uR++) {
	      index_map[*uR] = newIndex;
	      newIndex++;
	  }
	  for(int i = 0; i < local_restriction.dim(); i++) {
	      new_local_restriction |= attach_operation(local_restriction[i] * usedRays, pm::operations::associative_access<Map<int,int>,int>(&index_map));
	  }
      }
      
      
      
      newrays = newrays.minor(usedRays,All);
      newMaximalCones = newMaximalCones.minor(All,usedRays);
    }
    
    perl::Object result("WeightedComplex");
      result.take("RAYS") << newrays;
      result.take("MAXIMAL_CONES") << newMaximalCones;
      result.take("USES_HOMOGENEOUS_C") << uses_homog;
      result.take("LINEALITY_SPACE") << lineality;
      result.take("LOCAL_RESTRICTION") << new_local_restriction;
   
    return result;
  }
  
  // ------------------------- PERL WRAPPERS ---------------------------------------------------
  
  Function4perl(&separateRayMatrix,"separateRayMatrix(Matrix<Rational>,$)");
  
  Function4perl(&compute_product_complex,"compute_product_complex(WeightedComplex+)");
  
  Function4perl(&compute_product_complex_lattice, "compute_product_complex_lattice(WeightedComplex+)");
  
  Function4perl(&facetRefinement,"facetRefinement(WeightedComplex,Matrix<Rational>)");
  
  
  UserFunction4perl("# @category Basic polyhedral operations"
		    "# Takes a polyhedral complex and applies an affine linear transformation, given by a"
		    "# translate vector and a matrix. The method assumes the function is bijective (on the "
		    "# complex) and preserves cones, i.e. it just applies the transformation to the rays and "
		    "# lineality space and leaves the cones and weights unchanged."
		    "# @param WeightedComplex complex The complex to be transformed"
		    "# @param Vector<Rational> translate A vector whose dimension should be equal to the column "
		    "# dimension of the transformation matrix"
		    "# @param Matrix<Integer> matrix An integer matrix "
		    "# @return WeightedComplex The transformed complex, in homogeneous coordinates",
    &affineTransformation, "affineTransformation(WeightedComplex, Vector<Rational>, Matrix<Integer>)");
  
  Function4perl(&skeleton_complex,"calculate_skeleton_complex(WeightedComplex,$;$=1)");
  
  UserFunction4perl("# @category Basic polyhedral operations"
		    "# Takes a polyhedral complex and returns a list of all the local vertex fans, "
		    "# i.e. for each affine ray r, the list contains the fan Star_complex(r) "
		    "# (in non-homogeneous coordinates)"
		    "# If the complex has a non-trivial [[LOCAL_RESTRICTION]], only the local fans"
		    "# at compatible vertices are computed"
		    "# @param WeightedComplex complex A tropical variety"
		    "# @return perl::ListReturn A list of WeightedComplex objects in "
		    "# non-homogeneous coordinates. The i-th complex corresponds to the i-th "
		    "# affine ray ( vertex). If the complex is not in homogeneous coordinates, "
		    "# the list contains just the complex itself ",
    &fan_decomposition,"fan_decomposition(WeightedComplex)");
  
  
}}